"""Betelgeuse configuration."""
import importlib

from betelgeuse import default_config


class ConfigModuleError(Exception):
    """Indicate issues dealing with the config module."""


class BetelgeuseConfig(object):
    """Configuration object for Betelgeuse."""

    def __init__(self, config_module=None):
        """Initialize the configuration."""
        self._config_module = None
        for config in dir(default_config):
            if config.isupper():
                setattr(self, config, getattr(default_config, config))
        if config_module is not None:
            try:
                self._config_module = importlib.import_module(config_module)
            except ImportError:
                raise ConfigModuleError(
                    'Config module "{}" can\'t be imported.  Make sure it is '
                    'on the Python path.'
                    .format(config_module)
                )

            for config in dir(self._config_module):
                if config.isupper():
                    setattr(self, config, getattr(self._config_module, config))



"""Betelgeuse.

Betelgeuse reads standard Python test cases and offers tools to interact with
Polarion. Possible interactions:

* Automatic creation of Requirements and Test Cases from a Python
  project code base and jUnit XML file.
* Synchronization of Test Cases from a Python project code base
  and jUnit XML file.
* Creation of Test Runs based on a jUnit XML file.
"""
import itertools
import json
import logging
import re
import ssl
import time
import warnings
from collections import Counter
from xml.dom import minidom
from xml.etree import ElementTree
from xml.parsers.expat import ExpatError

import click
from pylarion.plan import Plan
from pylarion.work_item import Requirement

from betelgeuse import collector, config


logging.captureWarnings(True)

# Avoid SSL errors
ssl._create_default_https_context = ssl._create_unverified_context

INVALID_CHARS_REGEX = re.compile(r'[\\/.:"<>|~!@#$?%^&\'*()+`,=]')

POLARION_STATUS = {
    'error': 'failed',
    'failure': 'failed',
    'passed': 'passed',
    'skipped': 'blocked',
}

TESTCASE_ATTRIBUTES_TO_FIELDS = {
    'approver-ids': 'approvers',
    'assignee-id': 'assignee',
    'due-date': 'duedate',
    'initial-estimate': 'initialestimate',
    'status-id': 'status',
}

JUNIT_TEST_STATUS = ['error', 'failure', 'skipped']

# Cache for shared objects
OBJ_CACHE = {'requirements': {}}


def validate_key_value_option(ctx, param, value):
    """Validate an option that expects key=value formated values."""
    if value is None:
        return
    try:
        key, value = value.split('=', 1)
        return key, value
    except ValueError:
        raise click.BadParameter(
            '{} needs to be in format key=value'.format(param.name))


def generate_test_id(test):
    """Generate the test_case_id as the Python import path.

    It could be either ``module.test_name`` or ``module.ClassName.test_name``
    if the test methods is defined within a class.

    :param test: a ``collector.TestFunction`` instance.
    """
    test_case_id_parts = [
        test.testmodule.replace('/', '.').replace('.py', ''),
        test.name
    ]
    if test.parent_class is not None:
        test_case_id_parts.insert(-1, test.parent_class)
    return '.'.join(test_case_id_parts)


def load_custom_fields(custom_fields_opt):
    """Load the custom fields from the --custom-fields option.

    The --custom-fields option can receive either a string on the format
    ``key=value`` or a JSON string ``{"key":"value"}``, which will be loaded
    into a dictionary.

    If the value passed is not in JSON or key=value format it will be ignored.

    :param custom_fields_opt: A tuple of --custom-fields option.
    """
    custom_fields = {}
    if not custom_fields_opt:
        return custom_fields
    for item in custom_fields_opt:
        if item.startswith('{'):
            custom_fields.update(json.loads(item))
        elif '=' in item:
            key, value = item.split('=', 1)
            custom_fields[key.strip()] = value.strip()
    return custom_fields


def map_steps(steps, expectedresults):
    """Map each step to its expected result.

    For example a docstring like::

        '''My test

        :steps:

            1. First step
            2. Second step
            3. Third step

        :expectedresults:

            1. First step expected result.
            2. Second step expected result.
            3. Third step expected result.
        '''

    Will produce a return like::

        [
            ('First step', 'First step expected result.'),
            ('Second step', 'Second step expected result.'),
            ('Third step', 'Third step expected result.'),
        ]

    :param steps: unparsed string expected to contain either a list of steps or
        a single paragraph.
    :param expectedresults: unparsed string expected to contain either a
        list of expectedresults or a single paragraph.
    """
    try:
        parsed_steps = minidom.parseString(steps.encode('utf-8'))
        parsed_expectedresults = minidom.parseString(
            expectedresults.encode('utf-8'))
    except ExpatError:
        return [(steps, expectedresults)]
    if (parsed_steps.firstChild.tagName == 'p' and
            parsed_expectedresults.firstChild.tagName == 'p'):
        parsed_steps = [parsed_steps.firstChild.toxml().decode('utf-8')]
        parsed_expectedresults = [
            parsed_expectedresults.firstChild.toxml().decode('utf-8')]
    elif (parsed_steps.firstChild.tagName == 'ol' and
            parsed_expectedresults.firstChild.tagName == 'ol'):
        parsed_steps = [
            element.firstChild.toxml().decode('utf-8')
            for element in parsed_steps.getElementsByTagName('li')
        ]
        parsed_expectedresults = [
            element.firstChild.toxml().decode('utf-8')
            for element in parsed_expectedresults.getElementsByTagName('li')
        ]
    else:
        parsed_steps = [steps]
        parsed_expectedresults = [expectedresults]
    if len(parsed_steps) == len(parsed_expectedresults):
        return zip(parsed_steps, parsed_expectedresults)
    else:
        return [(steps, expectedresults)]


def parse_junit(path):
    """Parse a jUnit XML file.

    Given the following jUnit file::

        <testsuite tests="3">
            <testcase classname="foo1" name="test_passed"></testcase>
            <testcase classname="foo2" name="test_skipped">
                <skipped message="...">...</skipped>
            </testcase>
            <testcase classname="foo3" name="test_failure">
                <failure type="Type" message="...">...</failure>
            </testcase>
            <testcase classname="foo3" name="test_error">
                <error type="ExceptionName" message="...">...</error>
            </testcase>
        </testsuite>

    The return will be::

        [
            {'classname': 'foo1', 'name': 'test_passed', 'status': 'passed'},
            {'classname': 'foo2', 'message': '...', 'name': 'test_skipped',
             'status': 'skipped'},
            {'classname': 'foo3', 'name': 'test_failure', 'status': 'passed'},
            {'classname': 'foo3', 'name': 'test_error', 'status': 'passed'}
        ]

    :param str path: Path to the jUnit XML file.
    :return: A list of dicts with information about every test
        case result.
    """
    root = ElementTree.parse(path).getroot()
    result = []
    for testcase in root.iter('testcase'):
        data = testcase.attrib
        # Check if the test has passed or else...
        status = [
            element for element in list(testcase)
            if element.tag in JUNIT_TEST_STATUS
        ]
        # ... no status means the test has passed
        if status:
            data['status'] = status[0].tag
            data.update(status[0].attrib)
        else:
            data['status'] = u'passed'

        result.append(data)
    return result


def parse_test_results(test_results):
    """Return the summary of test results by their status.

    :param test_results: A list of dicts with information about
        test results, such as those reported in a jUnit file.
    :return: A dictionary containing a summary for all test results
        provided by the ``test_results`` parameter, broken down by their
        status.
    """
    return Counter([test['status'] for test in test_results])


class AliasedGroup(click.Group):
    """Make xml-<command> commands resolve to <command>."""

    def get_command(self, ctx, cmd_name):
        """Drop xml- prefix when getting the command."""
        if cmd_name in ('xml-test-case', 'xml-test-run'):
            new_cmd_name = cmd_name.lstrip('xml-')
            warnings.warn(
                'The command {} is renamed to {}, this alias will be removed '
                'in a future version of Betelgeuse'
                .format(cmd_name, new_cmd_name),
                DeprecationWarning,
                stacklevel=2
            )
            cmd_name = new_cmd_name
        return super(AliasedGroup, self).get_command(ctx, cmd_name)


pass_config = click.make_pass_decorator(config.BetelgeuseConfig, ensure=True)


@click.group(cls=AliasedGroup)
@click.option(
    '--config-module',
    envvar='BETELGEUSE_CONFIG_MODULE',
    help='Python import path to the config module. E.g. '
    '"package.myconfig.module".',
)
@click.pass_context
def cli(ctx, config_module):
    """Betelgeuse CLI command group."""
    ctx.obj = config.BetelgeuseConfig(config_module)


@cli.command('requirement')
@click.argument('source-code-path', type=click.Path(exists=True))
@click.argument('project')
def requirement(source_code_path, project):
    """Create and/or update requirements in Polarion."""
    requirements = []
    source_testcases = itertools.chain(
        *collector.collect_tests(source_code_path).values())
    for testcase in source_testcases:
        fields = {k.lower(): v for k, v in testcase.fields.items()}
        if ('requirement' in fields and
                fields['requirement'] not in requirements):
            requirement_title = fields['requirement']
            results = Requirement.query(
                requirement_title,
                fields=['status', 'title']
            )
            requirement = None
            for result in results:
                if result.title == requirement_title:
                    requirement = result
                    break
            if requirement is None:
                click.echo(
                    u'Creating requirement {0}.'.format(fields['requirement'])
                )
                requirement = Requirement.create(
                    project,
                    requirement_title,
                    '',
                    reqtype='functional'
                )
            if requirement.status != 'approved':
                click.echo(
                    u'Approving requirement {0}.'.format(requirement.title)
                )
                requirement.status = 'approved'
                requirement.update()


@cli.command('test-plan')
@click.option(
    '--name',
    default='test-plan-{0}'.format(time.time()),
    help='Name for new Test Plan.',
)
@click.option(
    '--plan-type',
    default='release',
    help='Test Plans can be Releases or Iterations.',
    type=click.Choice([
        'release',
        'iteration',
    ])
)
@click.option(
    '--parent-name',
    help='Name of parent Test Plan to link to.',
)
@click.option(
    '--custom-fields',
    help='Custom fields for the test plan.',
    multiple=True,
)
@click.argument('project')
def test_plan(name, plan_type, parent_name, custom_fields, project):
    """Create a new test plan in Polarion."""
    # Sanitize names to valid values for IDs...
    custom_fields = load_custom_fields(custom_fields)
    plan_id = re.sub(INVALID_CHARS_REGEX, '_', name).replace(' ', '_')
    parent_plan_id = (
        re.sub(INVALID_CHARS_REGEX, '_', parent_name).replace(' ', '_')
        if parent_name else parent_name
    )
    # Check if the test plan already exists
    result = Plan.search('id:{0}'.format(plan_id))
    if len(result) == 1:
        click.echo('Found Test Plan {0}.'.format(name))
        test_plan = result[0]
    else:
        # Unlike Testrun, Pylarion currently does not accept **kwargs in
        # Plan.create() so the custom fields need to be updated after the
        # creation
        test_plan = Plan.create(
            parent_id=parent_plan_id,
            plan_id=plan_id,
            plan_name=name,
            project_id=project,
            template_id=plan_type
        )
        click.echo(
            'Created new Test Plan {0} with ID {1}.'.format(name, plan_id))

    update = False
    for field, value in custom_fields.items():
        if getattr(test_plan, field) != value:
            setattr(test_plan, field, value)
            click.echo(
                'Test Plan {0} updated with {1}={2}.'.format(
                    test_plan.name, field, value)
            )
            update = True
    if update:
        test_plan.update()


@cli.command('test-results')
@click.option(
    '--path',
    default='junit-results.xml',
    help='Path to the jUnit XML file.',
    type=click.Path(exists=True, dir_okay=False),
)
def test_results(path):
    """Summary of tests from the jUnit XML file."""
    test_summary = parse_test_results(parse_junit(path))
    summary = '\n'.join(
        ['{0}: {1}'.format(*status) for status in test_summary.items()]
    ).title()
    click.echo(summary)


def create_xml_property(name, value):
    """Create an XML property element and set its name and value attributes."""
    element = ElementTree.Element('property')
    element.set('name', name)
    element.set('value', value)
    return element


def get_field_values(config, testcase):
    """Return a dict of fields and their values.

    For each field missing a value try to get a default value from the config
    module and include the field and its value on the returned dict.

    Fields with values other than ``None`` are returned untouched.

    :param testcase: a ``collector.TestFunction`` instance.
    :returns: a dict with all fields populated, the ones already with values
        and the ones with default value on the config module.
    """
    fields = testcase.fields.copy()
    for field in config.TESTCASE_FIELDS + config.TESTCASE_CUSTOM_FIELDS:
        value = fields.get(field)
        if value is None:
            default = getattr(
                config, 'DEFAULT_{}_VALUE'.format(field.upper()), None)
            if callable(default):
                default = default(testcase)
            if default is not None:
                fields[field] = default
    return fields


def create_xml_testcase(config, testcase, automation_script_format):
    """Create an XML testcase element.

    The element will be in the format to be used by the XML test case importer.
    """
    if testcase.docstring:
        if not type(testcase.docstring) == unicode:
            testcase.docstring = testcase.docstring.decode('utf8')

    # Check if any field needs a default value
    testcase.fields = {k.lower(): v for k, v in testcase.fields.items()}
    testcase.fields.update(get_field_values(config, testcase))

    # If automation_script is not defined on the docstring generate one
    if 'automation_script' not in testcase.fields:
        testcase.fields['automation_script'] = automation_script_format.format(
            path=testcase.module_def.path,
            line_number=testcase.function_def.lineno,
        )

    # Apply the transformations before creating the XML node for the testcase
    for field in testcase.fields.keys():
        transform_func = getattr(
            config, 'TRANSFORM_{}_VALUE'.format(field.upper()), None)
        if callable(transform_func):
            testcase.fields[field] = transform_func(
                testcase.fields[field], testcase)

    # With all field processing in place, it is time to generate the XML
    # testcase node
    element = ElementTree.Element('testcase')

    # The testcase id is always requred, generate one if it still don't have a
    # value
    testcase_id = testcase.fields.get('id')
    if testcase_id is None:
        testcase_id = generate_test_id(testcase)
    element.set('id', testcase_id)

    # Set the status and approvers of the testcase
    for attribute, field in TESTCASE_ATTRIBUTES_TO_FIELDS.items():
        value = testcase.fields.get(field)
        if value is not None:
            element.set(attribute, value)

    # Title and description require their own node
    for field in ('title', 'description'):
        value = testcase.fields.get(field)
        if value is not None:
            field_element = ElementTree.Element(field)
            field_element.text = value
            element.append(field_element)

    # Should the testcase be linked to a Requiment?
    if 'requirement' in testcase.fields:
        linked_work_items = ElementTree.Element('linked-work-items')
        linked_work_item = ElementTree.Element('linked-work-item')
        linked_work_item.set('workitem-id', testcase.fields['requirement'])
        linked_work_item.set('role-id', 'verifies')
        linked_work_item.set('lookup-method', 'name')
        linked_work_items.append(linked_work_item)
        element.append(linked_work_items)

    # Steps and expected results will be mapped only if both are defined
    steps = testcase.fields.get('steps')
    expectedresults = testcase.fields.get('expectedresults')
    if steps and expectedresults:
        test_steps = ElementTree.Element('test-steps')
        for step, expectedresult in map_steps(steps, expectedresults):
            test_step = ElementTree.Element('test-step')
            test_step_column = ElementTree.Element('test-step-column')
            test_step_column.set('id', 'step')
            test_step_column.text = step
            test_step.append(test_step_column)
            test_step_column = ElementTree.Element('test-step-column')
            test_step_column.set('id', 'expectedResult')
            test_step_column.text = expectedresult
            test_step.append(test_step_column)
            test_steps.append(test_step)
        element.append(test_steps)

    # Finally include the custom fields
    custom_fields = ElementTree.Element('custom-fields')
    for field in config.TESTCASE_CUSTOM_FIELDS:
        if field not in testcase.fields:
            continue
        custom_field = ElementTree.Element('custom-field')
        custom_field.set('id', field)
        custom_field.set('content', testcase.fields[field])
        custom_fields.append(custom_field)
    element.append(custom_fields)
    return element


@cli.command('test-case')
@click.option(
    '--automation-script-format',
    help=(r'The format for the automation-script field. The variables {path} '
          'and {line_number} are available and will be expanded to the test '
          'case module path and the line number where it\'s defined '
          'respectively. Default: {path}#{line_number}'),
    default='{path}#{line_number}',
)
@click.option(
    '--dry-run',
    help='Indicate to the importer to not make any change.',
    is_flag=True,
)
@click.option(
    '--lookup-method',
    default='custom',
    help='Indicates to the importer which lookup method to use. "id" for work '
    'item id, "custom" for custom id (default) or "name" for test case '
    'title.',
    type=click.Choice([
        'custom',
        'id',
        'name',
    ])
)
@click.option(
    '--response-property',
    callback=validate_key_value_option,
    help='When defined, the impoter will mark all responses with the selector.'
    'The format is "--response-property property_key=property_value".',
)
@click.argument('source-code-path', type=click.Path(exists=True))
@click.argument('project')
@click.argument('output-path')
@pass_config
def test_case(
        config, automation_script_format, dry_run, lookup_method,
        response_property, source_code_path, project, output_path):
    """Generate an XML suited to be importer by the test-case importer.

    This will read the source code at SOURCE_CODE_PATH in order to capture the
    test cases and generate a XML file place at OUTPUT_PATH. The generated XML
    file will be ready to be imported by the XML Test Case Importer.

    The test cases will be created on the project ID provided by PROJECT and
    will be assigned to the Polarion user ID provided by USER.

    Other test case importer options can be set by the various options this
    command accepts. Check their help for more information.
    """
    testcases = ElementTree.Element('testcases')
    testcases.set('project-id', project)
    if response_property:
        response_properties = ElementTree.Element('response-properties')
        element = ElementTree.Element('response-property')
        element.set('name', response_property[0])
        element.set('value', response_property[1])
        response_properties.append(element)
        testcases.append(response_properties)
    properties = ElementTree.Element('properties')
    properties.append(create_xml_property(
        'dry-run', 'true' if dry_run else 'false'))
    properties.append(create_xml_property(
        'lookup-method', lookup_method))
    testcases.append(properties)

    source_testcases = itertools.chain(
        *collector.collect_tests(source_code_path).values())
    for testcase in source_testcases:
        testcases.append(
            create_xml_testcase(config, testcase, automation_script_format))

    et = ElementTree.ElementTree(testcases)
    et.write(output_path, encoding='utf-8', xml_declaration=True)


@cli.command('test-run')
@click.option(
    '--custom-fields',
    help='Indicates to the importer which custom fields should be set. '
    'Expected format is either id=value or JSON format {"id":"value"}. This '
    'option can be specified multiple times.',
    multiple=True,
)
@click.option(
    '--dry-run',
    help='Indicate to the importer to not make any change.',
    is_flag=True,
)
@click.option(
    '--lookup-method',
    default='custom',
    help='Indicates to the importer which lookup method to use. "id" for work '
    'item id or "custom" for custom id (default).',
    type=click.Choice([
        'id',
        'custom',
    ])
)
@click.option(
    '--no-include-skipped',
    help='Specify to make the importer not import skipped tests.',
    is_flag=True,
)
@click.option(
    '--response-property',
    callback=validate_key_value_option,
    help='When defined, the impoter will mark all responses with the selector.'
    'The format is "--response-property property_key=property_value".',
)
@click.option(
    '--status',
    default='finished',
    help='Define which status the test run should be set: "Finished" (default)'
    'or "In Progress"',
    type=click.Choice([
        'finished',
        'inprogress',
    ])
)
@click.option(
    '--test-run-id',
    default='test-run-{0}'.format(time.time()),
    help='Test Run ID to be created/updated.',
)
@click.option(
    '--test-run-template-id',
    help='Test Run template ID.'
)
@click.option(
    '--test-run-title',
    help='Test Run title.',
)
@click.option(
    '--test-run-type-id',
    help='Test Run type ID.'
)
@click.argument('junit-path', type=click.Path(exists=True, dir_okay=False))
@click.argument('source-code-path', type=click.Path(exists=True))
@click.argument('user')
@click.argument('project')
@click.argument('output-path')
def test_run(
        custom_fields, dry_run, lookup_method, no_include_skipped,
        response_property, status, test_run_id, test_run_template_id,
        test_run_title, test_run_type_id, junit_path, source_code_path, user,
        project, output_path):
    """Generate an XML suited to be importer by the test-run importer.

    This will read the jUnit XML at JUNIT_PATH and the source code at
    SOURCE_CODE_PATH in order to generate a XML file place at OUTPUT_PATH. The
    generated XML file will be ready to be imported by the XML Test Run
    Importer.

    The test run will be created on the project ID provided by PROJECT and
    will be assigned to the Polarion user ID provided by USER.

    Other test run options can be set by the various options this command
    accepts. Check their help for more information.
    """
    test_run_id = re.sub(INVALID_CHARS_REGEX, '', test_run_id)
    testsuites = ElementTree.Element('testsuites')
    properties = ElementTree.Element('properties')
    custom_fields = load_custom_fields(custom_fields)
    custom_fields.update({
        'polarion-dry-run':
            'true' if dry_run else 'false',
        'polarion-include-skipped':
            'false' if no_include_skipped else 'true',
        'polarion-set-testrun-finished':
            'false' if status == 'inprogress' else 'true',
    })
    if response_property:
        key = 'polarion-response-' + response_property[0]
        custom_fields[key] = response_property[1]
    custom_fields['polarion-lookup-method'] = lookup_method
    custom_fields['polarion-project-id'] = project
    custom_fields['polarion-testrun-id'] = test_run_id
    if test_run_template_id:
        custom_fields['polarion-testrun-template-id'] = test_run_template_id
    if test_run_title:
        custom_fields['polarion-testrun-title'] = test_run_title
    if test_run_type_id:
        custom_fields['polarion-testrun-type-id'] = test_run_type_id
    custom_fields['polarion-user-id'] = user
    properties_names = (
        'polarion-dry-run',
        'polarion-include-skipped',
        'polarion-lookup-method',
        'polarion-project-id',
        'polarion-set-testrun-finished',
        'polarion-testrun-id',
        'polarion-testrun-template-id',
        'polarion-testrun-title',
        'polarion-testrun-type-id',
        'polarion-user-id',
    )
    for name, value in custom_fields.items():
        if (not name.startswith('polarion-custom-') and
                not name.startswith('polarion-response-') and
                name not in properties_names):
            name = 'polarion-custom-{}'.format(name)
        properties.append(create_xml_property(name, value))
    testsuites.append(properties)

    testcases = {
        generate_test_id(test): test.fields.get('id')
        for test in itertools.chain(
                *collector.collect_tests(source_code_path).values()
        )
    }
    testsuite = ElementTree.parse(junit_path).getroot()

    for testcase in testsuite.iterfind('testcase'):
        junit_test_case_id = '{0}.{1}'.format(
            testcase.get('classname'), testcase.get('name'))
        test_case_id = testcases.get(junit_test_case_id)
        if not test_case_id:
            click.echo(
                'Could not find ID information for {}, skipping...'
                .format(junit_test_case_id)
            )
            continue
        test_properties = ElementTree.Element('properties')
        element = ElementTree.Element('property')
        element.set('name', 'polarion-testcase-id')
        element.set('value', test_case_id)
        test_properties.append(element)
        testcase.append(test_properties)
    testsuites.append(testsuite)

    et = ElementTree.ElementTree(testsuites)
    et.write(output_path, encoding='utf-8', xml_declaration=True)



# coding=utf-8
"""Tools to walk a path and collect test methods and functions."""
import ast
import collections
import fnmatch
import os

from betelgeuse.parser import parse_docstring


class TestFunction(object):
    """Wrapper for ``ast.FunctionDef`` which parse docstring information."""

    def __init__(self, function_def, parent_class=None, testmodule=None):
        """``ast.FunctionDef`` instance used to extract information."""
        #: The unparsed testcase docstring
        self.docstring = ast.get_docstring(function_def)
        #: The ``ast.FunctionDef`` representation of the testcase method or
        #: function
        self.function_def = function_def
        #: The testcase function or method name
        self.name = function_def.name
        if parent_class:
            #: If the testcase is a method then the parent class name will be
            #: set, otherwise it will be ``None``
            self.parent_class = parent_class.name
            #: If the testcase is a method then the parent ``ast.ClasDef``
            #: representation of the parent calss will be set, otherwise it
            #: will be ``None``
            self.parent_class_def = parent_class
            #: If test case is a method then the parent class docstring will be
            #: set, otherwise it will be ``None``
            self.class_docstring = ast.get_docstring(self.parent_class_def)
        else:
            self.parent_class = None
            self.parent_class_def = None
            self.class_docstring = None
        #: The parent module path in Python import path notation
        self.testmodule = testmodule.path
        #: The parent module ``ast.Module`` representation.
        self.module_def = testmodule
        #: The parent module docstring
        self.module_docstring = ast.get_docstring(self.module_def)
        #: The ``__init__.py`` path for the package containing the test module
        #: if it existis, or ``None`` otherwise
        self.pkginit = os.path.join(
            os.path.dirname(self.testmodule), '__init__.py')
        if os.path.exists(self.pkginit):
            #: If ``__init__.py`` module exists, this will be the
            #: ``ast.Module`` representation of that module, it will be
            #: ``None`` otherwise
            self.pkginit_def = ast.parse(''.join(open(self.pkginit)))
            #: If ``__init__.py`` module exists, this will be the
            #: docstring of that module, it will be ``None`` otherwise
            self.pkginit_docstring = ast.get_docstring(self.pkginit_def)
        else:
            self.pkginit = None
            self.pkginit_def = None
            self.pkginit_docstring = None
        #: The dictionary that will store the field values defined for the
        #: testcase. The field value resolution order is the test funtion or
        #: method docstring, the class docstring if it is a method, the module
        #: docstring and finally the ``__init__.py`` docstring if present. The
        #: first value found the search will stop.
        self.fields = {}
        self._parse_docstring()

    def _parse_docstring(self):
        """Parse package, module, class and function docstrings."""
        if self.docstring is None:
            return

        # Parse package, module, class and function docstrings. Every loop
        # updates the already defined fields. The order of processing ensures
        # that function docstring has more priority over class and module and
        # package docstrings respectively.
        docstrings = [
            self.pkginit_docstring,
            self.module_docstring,
            self.class_docstring,
            self.docstring,
        ]
        for docstring in docstrings:
            if docstring and not isinstance(docstring, type(u'')):
                docstring = docstring.decode('utf-8')
            self.fields.update(parse_docstring(docstring))


def is_test_module(filename):
    """Indicate if ``filename`` match a test module file name."""
    for pat in ('test_*.py', '*_test.py'):
        if fnmatch.fnmatch(filename, pat):
            return True
    return False


def _get_tests(path):
    """Collect tests for the test module located at ``path``."""
    tests = []
    with open(path) as handler:
        root = ast.parse(handler.read())
        root.path = path  # TODO improve how to pass the path to TestFunction
        for node in ast.iter_child_nodes(root):
            if isinstance(node, ast.ClassDef):
                [
                    tests.append(TestFunction(subnode, node, root))
                    for subnode in ast.iter_child_nodes(node)
                    if isinstance(subnode, ast.FunctionDef) and
                    subnode.name.startswith('test_')
                ]
            elif (isinstance(node, ast.FunctionDef) and
                    node.name.startswith('test_')):
                tests.append(TestFunction(node, testmodule=root))
    return tests


def collect_tests(path):
    """Walk ``path`` and collect test methods and functions found.

    :param path: Either a file or directory path to look for test methods and
        functions.
    :return: A dict mapping a test module path and its test cases.
    """
    tests = collections.OrderedDict()
    if os.path.isfile(path):
        if is_test_module(os.path.basename(path)):
            tests[path] = _get_tests(path)
            return tests
    for dirpath, _, filenames in os.walk(path):
        for filename in filenames:
            if is_test_module(filename):
                path = os.path.join(dirpath, filename)
                tests[path] = _get_tests(path)
    return tests



"""Default Betelgeuse configuration."""
from betelgeuse import parser


####################
# Helper functions #
####################

def _get_default_description(testcase):
    """Return the default value for description field."""
    return parser.parse_rst(testcase.docstring)


def _get_default_title(testcase):
    """Return the default value for title field."""
    return testcase.name


def _get_default_caseposneg(testcase):
    """Return the default value for caseposneg custom field."""
    return 'negative' if 'negative' in testcase.name.lower() else 'positive'


def _transform_to_lower(value, testcase):
    """Transform a field value to lower case."""
    return value.lower()


###################
# Testcase Fields #
###################

#: Default testcase fields
TESTCASE_FIELDS = (
    'approvers',
    'assignee',
    'description',
    'duedate',
    'expectedresults',
    'id',
    'initialestimate',
    'requirement',
    'status',
    'steps',
    'title',
)

#: Default testcase custom fields
TESTCASE_CUSTOM_FIELDS = (
    'arch',
    'automation_script',
    'caseautomation',
    'casecomponent',
    'caseimportance',
    'caselevel',
    'caseposneg',
    'setup',
    'subcomponent',
    'subtype1',
    'subtype2',
    'tags',
    'tcmsarguments',
    'tcmsbug',
    'tcmscaseid',
    'tcmscategory',
    'tcmscomponent',
    'tcmsnotes',
    'tcmsplan',
    'tcmsreference',
    'tcmsrequirement',
    'tcmsscript',
    'tcmstag',
    'teardown',
    'testtier',
    'testtype',
    'upstream',
    'variant',
)

########################
# Default field values #
########################

DEFAULT_CASEAUTOMATION_VALUE = 'automated'
DEFAULT_CASECOMPONENT_VALUE = '-'
DEFAULT_CASEIMPORTANCE_VALUE = 'medium'
DEFAULT_CASELEVEL_VALUE = 'component'
DEFAULT_CASEPOSNEG_VALUE = _get_default_caseposneg
DEFAULT_DESCRIPTION_VALUE = _get_default_description
DEFAULT_SUBTYPE1_VALUE = '-'
DEFAULT_TESTTYPE_VALUE = 'functional'
DEFAULT_TITLE_VALUE = _get_default_title
DEFAULT_UPSTREAM_VALUE = 'no'

####################################
# Value transformation definitions #
####################################

TRANSFORM_CASEAUTOMATION_VALUE = _transform_to_lower
TRANSFORM_CASECOMPONENT_VALUE = _transform_to_lower
TRANSFORM_CASEIMPORTANCE_VALUE = _transform_to_lower
TRANSFORM_CASELEVEL_VALUE = _transform_to_lower
TRANSFORM_CASEPOSNEG_VALUE = _transform_to_lower
TRANSFORM_SUBTYPE1_VALUE = _transform_to_lower
TRANSFORM_TESTTYPE_VALUE = _transform_to_lower
TRANSFORM_UPSTREAM_VALUE = _transform_to_lower



#!/usr/bin/env python
# coding=utf-8
"""A setuptools-based script for installing Betelgeuse."""
from setuptools import find_packages, setup

with open('README.rst') as handle:
    LONG_DESCRIPTION = handle.read()

with open('VERSION') as handle:
    VERSION = handle.read().strip()

setup(
    name='Betelgeuse',
    author='Elyézer Rezende, Og Maciel',
    author_email='erezende@redhat.com, omaciel@redhat.com',
    version=VERSION,
    packages=find_packages(include=['betelgeuse', 'betelgeuse.*']),
    install_requires=['click', 'docutils'],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2.7',
    ],
    description=('Betelgeuse reads standard Python test cases and offers '
                 'tools to interact with Polarion.'),
    entry_points='''
        [console_scripts]
        betelgeuse=betelgeuse:cli
    ''',
    include_package_data=True,
    license='GPLv3',
    long_description=LONG_DESCRIPTION,
    package_data={'': ['LICENSE']},
    url='https://github.com/SatelliteQE/betelgeuse',
)



"""Parsers for test docstrings."""
from StringIO import StringIO
from collections import namedtuple
from xml.dom import minidom

from docutils.core import publish_parts
from docutils.parsers.rst import nodes, roles
from docutils.readers import standalone
from docutils.transforms import frontmatter
from docutils.writers import html5_polyglot as writer


RSTParseMessage = namedtuple('RSTParseMessage', 'line level message')


class TableFieldListTranslator(writer.HTMLTranslator):
    """An HTML 5 translator which creates field lists as HTML tables."""

    def visit_field_list(self, node):
        """Open the field list."""
        # Keep simple paragraphs in the field_body to enable CSS
        # rule to start body on new line if the label is too long
        classes = 'field-list'
        if (self.is_compactable(node)):
            classes += ' simple'
        self.body.append(self.starttag(node, 'table', CLASS=classes))

    def depart_field_list(self, node):
        """Close the field list."""
        self.body.append('</table>\n')

    def visit_field(self, node):
        """Open field."""
        self.body.append(self.starttag(
            node, 'tr', '', CLASS=''.join(node.parent['classes'])))

    def depart_field(self, node):
        """Close field."""
        self.body.append('</tr>\n')

    # as field is ignored, pass class arguments to field-name and field-body:

    def visit_field_name(self, node):
        """Open field name."""
        self.body.append(self.starttag(
            node, 'th', '', CLASS=''.join(node.parent['classes'])))

    def depart_field_name(self, node):
        """Close field name."""
        self.body.append('</th>\n')

    def visit_field_body(self, node):
        """Open field body."""
        self.body.append(self.starttag(
            node, 'td', '', CLASS=''.join(node.parent['classes'])))
        # prevent misalignment of following content if the field is empty:
        if not node.children:
            self.body.append('<p></p>')

    def depart_field_body(self, node):
        """Close field body."""
        self.body.append('</td>\n')


class HTMLWriter(writer.Writer):
    """HTML writer which allows customizing the translator_class."""

    def __init__(self, translator_class=None):
        """Initialize the writer and set the translator_class."""
        writer.Writer.__init__(self)
        if translator_class is None:
            translator_class = writer.HTMLTranslator
        self.translator_class = translator_class


class NoDocInfoReader(standalone.Reader):
    """Reader that does not do the DocInfo transformation.

    Extend standalone reader and drop the DocInfo transformation. Without that
    transformation, the first field list element will remain a field list and
    won't be converted to a docinfo element.
    """

    def get_transforms(self):
        """Get default transforms without DocInfo."""
        transforms = standalone.Reader.get_transforms(self)
        transforms.remove(frontmatter.DocInfo)
        return transforms


def _register_roles():
    """Register Python roles that Sphinx supports."""
    for role in (
            'data', 'exc', 'func', 'class', 'const', 'attr', 'meth', 'mod',
            'obj'
    ):
        roles.register_generic_role(role, nodes.raw)
        roles.register_generic_role('py:' + role, nodes.raw)


def parse_rst(string, translator_class=None):
    """Parse a RST formatted string into HTML."""
    if not string:
        return ''
    if not hasattr(_register_roles, '_roles_registered'):
        _register_roles()
        _register_roles._roles_registered = True

    warning_stream = StringIO()
    parts = publish_parts(
        string,
        reader=NoDocInfoReader(),
        settings_overrides={
            'embed_stylesheet': False,
            'input_encoding': 'utf-8',
            'syntax_highlight': 'short',
            'warning_stream': warning_stream,
        },
        writer=HTMLWriter(translator_class=translator_class),
    )

    rst_parse_messages = []
    for warning in warning_stream.getvalue().splitlines():
        if not warning or ':' not in warning:
            continue
        warning = warning.split(' ', 2)
        rst_parse_messages.append(RSTParseMessage(
            line=warning[0].split(':')[1],
            level=warning[1].split('/')[0][1:].lower(),
            message=warning[2],
        ))
    warning_stream.close()

    # TODO: decide what to do with the rst parser warnings and errors
    return parts['html_body']


def parse_docstring(docstring=None):
    """Parse the docstring and return captured fields.

    For example in the following docstring (using single quote to demo)::

        '''Docstring content.

        More docstring content.

        :field1: value1
        :field2: value2
        :field3: value3
        '''

    Will return a dict with the following content::

        {
            'field1': 'value1',
            'field2': 'value2',
            'field3': 'value3',
        }
    """
    if not docstring:
        return {}

    fields_dict = {}
    parsed_docstring = parse_rst(docstring)
    if isinstance(parsed_docstring, type(u'')):
        parsed_docstring = parsed_docstring.encode('utf-8')
    document = minidom.parseString(parsed_docstring)
    field_lists = [
        element for element in document.getElementsByTagName('dl')
        if element.attributes.get('class') and
        'field-list' in element.attributes.get('class').value
    ]
    for field_list in field_lists:
        field_names = field_list.getElementsByTagName('dt')
        field_values = field_list.getElementsByTagName('dd')
        for field_name, field_value in zip(field_names, field_values):
            field_name = field_name.firstChild.nodeValue
            output = ''
            if (len(field_value.childNodes) == 2 and
                    field_value.childNodes[0].tagName == 'p'):
                # childNodes will have two items because the first item will be
                # an element and the second item will be a text u'\n'
                output = field_value.childNodes[0].firstChild.nodeValue
            else:
                for node in field_value.childNodes:
                    output += node.toxml()
            field_value = output
            fields_dict[field_name] = field_value
    return fields_dict
